#!python
# Copyright (c) 2000-2008 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys

from xpcom import components, nsError, ServerException, COMException

import eollib
import uriparse
import koprocessutils
from koAsyncOperationUtils import koAsyncOperationBase



#---- helper functions

def splitFile(file):
    """Split the filepath (or uri) into two parts, the directory and the
    filename. Returns a tuple of (dirName, fileName).
    """
    path = uriparse.URIToLocalPath(file);
    if path:
        if os.path.isdir(path):
            return path, ''
        path = os.path.normpath(path)
        basedir = os.path.dirname(path)
        filename = os.path.basename(path)
        return basedir, filename
    return None, None

def groupFilesByDirectory(files):
    """Split all the files (or can be uris) into directory and filename
    components, grouping all files with the same directory into one dictionary.
    Returns a dictionary whose keys are the distinct directories, with each
    key holding the list of files with this directory.
    """
    _files = {}
    for file in files:
        basedir, filename = splitFile(file)
        if basedir:
            if not basedir in _files:
                _files[basedir] = []
            if filename:
                _files[basedir].append(filename)
    return _files


#---- class to assist with file management calls.

class PathHelperMixin(object):

    def directoryExists(self, path):
        # Will not work for remote paths, override this if using remote paths.
        return os.path.isdir(path)

    def getParentDirectory(self, path):
        # May not work for correctly for remote paths, override this if using
        # remote paths.
        return os.path.dirname(path)

    def joinPaths(self, path1, path2):
        # May not work for correctly for remote paths, override this if using
        # remote paths.
        return os.path.join(path1, path2)

    def pathExists(self, path):
        # Will not work for remote paths, override this if using remote paths.
        return os.path.exists(path)

    def getParentDirContainingDirname(self, dirName, fromDir, max_traversal=20):
        dirPath = self.joinPaths(fromDir, dirName)
        dir_count = 1
        while not self.directoryExists(dirPath):
            lastDir = fromDir
            fromDir = self.getParentDirectory(fromDir)
            if not fromDir or lastDir == fromDir or dir_count >= max_traversal:
                # We did not find a parent directory with this name.
                return None
            dirPath = self.joinPaths(fromDir, dirName)
            dir_count += 1
        return fromDir


#---- component implementation

class KoSCCBase(PathHelperMixin):

    # Satisfy some of koISCC
    name = "xxx"
    reasonNotFunctional = ""
    isFunctional = False

        # executablePrefName is the preference name used to store the scc
        # command that gets used by the scc library calls.
    executablePrefName = "xxxExecutable"
        # executableBaseName is the fallback command that is used to set the
        # location of _executable and determine if the service is
        # functional. Example: "svn"
    executableBaseName = "xxx"

    def __init__(self):
        PathHelperMixin.__init__(self)
            # _executable is the command that will be executed by the scc
            # library.
        self._executable = None

        # Setup some common services the scc component will need to use.
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._encodingSvc = components.classes['@activestate.com/koEncodingServices;1'].\
                                getService(components.interfaces.koIEncodingServices)

    def _findExecutable(self, command):
        import which
        assert which._version_ >= (1,0,0)
        try:
            if sys.platform.startswith('win'):
                exts = ['.exe']
            else:
                exts = None
            env = koprocessutils.getUserEnv()
            userPath = env.get("PATH", "").split(os.pathsep)
            executable = which.which(command, exts=exts, path=userPath)
        except which.WhichError:
            executable = None
        return executable

    def set_executable(self, executable):
        if not executable:
            executable = self._findExecutable(self.executableBaseName)
        if executable != self._executable:
            self._executable = executable
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', self.name)
            except COMException:
                pass # Ignore error if there is no observer.

    def _setExecutableFromPrefs(self, prefName=None):
        executable = None
        if prefName is None:
            prefName = self.executablePrefName
        if self._globalPrefs.hasPref(prefName):
            executable = self._globalPrefs.getStringPref(prefName)
        self.set_executable(executable)

    def get_executable(self):
        if not self._executable:
            self._setExecutableFromPrefs()
        return self._executable

    def redetermineIfFunctional(self):
        executable = None
        wasFunctional = self.isFunctional
        if not self._executable:
            self.set_executable(None)  # Looks again for an executable.
        if self._executable and os.path.exists(self._executable):
            self.isFunctional = True
            self.reasonNotFunctional = None
        else:
            self.isFunctional = False
            self.reasonNotFunctional = "no %s executable could be found" % (self.name, )

        if wasFunctional != self.isFunctional:
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', self.name)
            except COMException:
                pass # Ignore error if there is no observer.

    def convertDiffResult(self, diff):
        # Convert line-endings to the currently expected
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)
        # Try encoding it with komodo's unicode encoding service
        diff, encoding, bom = self._encodingSvc.getUnicodeEncodedString(diff)
        return diff

# --- The asynchronous wrappers of the scc commands.


    def diff(self, files, options, external, async_callback):
        return self.runCommandAsynchronously("diff", self, self._do_diff,
                                             files, False, async_callback,
                                             files, options, external)

    # Diff two revisions of the given file
    def diffRevisions(self, fileuri1, rev1, fileuri2, rev2, localfilepath,
                      options, external, async_callback):
        return self.runCommandAsynchronously("diffRevisions", self,
                                             self._do_diff_revisions,
                                             [fileuri1, fileuri2], False,
                                             async_callback,
                                             fileuri1, rev1, fileuri2, rev2,
                                             localfilepath, options, external)

    def history(self, fileuri, options, async_callback):
        # Don't need to lock the files
        return self.runCommandAsynchronously("history", self, self._do_history,
                                             [fileuri], False, async_callback,
                                             fileuri, options)

    def add(self, files, mode, message, async_callback=None):
        return self.runCommandAsynchronously("add", self, self._do_add,
                                             files, False, async_callback,
                                             files, mode, message)

    def commit(self, files, message, options, async_callback=None):
        return self.runCommandAsynchronously("commit", self, self._do_commit,
                                             files, True, async_callback,
                                             files, message, options)

    def remove(self, files, force, recursive, async_callback):
        return self.runCommandAsynchronously("remove", self, self._do_remove,
                                             files, True, async_callback,
                                             files, force, recursive)
    
    def update(self, files, options, async_callback):
        return self.runCommandAsynchronously("update", self, self._do_update,
                                             files, False, async_callback,
                                             files, options)

    def revert(self, files, options, async_callback):
        return self.runCommandAsynchronously("revert", self, self._do_revert,
                                             files, True, async_callback,
                                             files, options)
    
    def edit(self, files, async_callback):
        return self.runCommandAsynchronously("edit", self, self._do_edit,
                                             files, False, async_callback,
                                             files)

    def status(self, files, recursive, options, async_callback):
        return self.runCommandAsynchronously("status", self, self._do_status,
                                             files, False, async_callback,
                                             files, recursive, options)

    def runCommandAsynchronously(self, cmd_name, sccSvc, run_function,
                                 affected_uris, lock_the_uris, async_callback,
                                 *args, **kwargs):
        # Run asynchronously
        async_svc = components.classes["@activestate.com/koAsyncService;1"].\
                        getService(components.interfaces.koIAsyncService)

        # Create a new instance to ensure we don't have multiple scc
        # operations running on the same instance.
        #scc_handler = svnlib.SVN()
        # Ensure the instance uses the same executable as the service.
        #scc_handler.executable = sccSvc.executable
        #kwargs['scc_handler'] = scc_handler

        async_op = koAsyncOperationBase(run_function, *args, **kwargs)
        async_svc.run("SCC %s %s" % (self.name, cmd_name),
                      async_op, async_callback, affected_uris, lock_the_uris)
        return async_op


# --- The actual implementations of the scc commands. Implement these for your
# --- individual scc component.


    def _do_diff(self, files, options, external, scc_handler=None):
        """Display diff of the client file with the repository file."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_diff_revisions(self, fileuri1, rev1, fileuri2, rev2, localfilepath,
                           options, external, scc_handler=None):
        """Display diff between two revisions of the one file."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_history(self, fileuri, options, scc_handler=None):
        """List revision history for this file."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_add(self, files, mode, message, scc_handler=None):
        """Open a new file to add it to the repository."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_commit(self, files, message, options, scc_handler=None):
        """Submit open files to the repository."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_remove(self, files, force, recursive, scc_handler=None):
        """Mark existing scc files to be deleted from the repository."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_update(self, files, options, scc_handler=None):
        """Synchronize the client with the repository."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_revert(self, files, options, scc_handler=None):
        """Discard changes from these opened files."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_edit(self, files, scc_handler=None):
        """Prepare the file for scc modification."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

    def _do_status(self, files, recursive, options, scc_handler=None):
        """Retrieve file scc information for these files."""
        raise ServerException(nsError.NS_ERROR_NOT_IMPLEMENTED)

